import binascii
import hmac
import os
import threading
from select import select
from socket import socket, AF_INET, SOCK_STREAM

from server.server.common.handler import Handler
from server.server.common.descripts import CheckPort, CheckIp
from server.server.common.metaclasses import CheckStatus
from login_required import login_required


class Server(threading.Thread, metaclass=CheckStatus):
    listen_port = CheckPort()
    ip_address = CheckIp()
    """
    The main server class. Accepts connections, packages
    from clients, processes incoming messages.
    """

    def __init__(self, ip_address, port, database):
        super().__init__()
        self.ip_address = ip_address
        self.listen_port = port
        self.database = database
        self.clients = []
        self.messages = None
        self.users = {}

    def run(self):
        """Main thread loop"""

        print(f'Address: {self.ip_address}\nPort: {self.listen_port}')

        sock = socket(AF_INET, SOCK_STREAM)
        sock.bind((self.ip_address, self.listen_port))
        sock.settimeout(0.5)

        sock.listen()

        while True:

            try:
                client, client_address = sock.accept()
            except OSError:
                pass
            else:
                print(f'The client {client_address} connected')
                self.clients.append(client)
            receive_data = []
            try:
                if self.clients:
                    receive_data, self.messages, errors = select(self.clients,
                                                                 self.clients,
                                                                 [], 0)
            except Exception as e:
                print(e)

            if receive_data:
                for clients_message in receive_data:
                    try:
                        handler = Handler()
                        handler.sock = clients_message
                        handler.get_message()
                        message = handler.message
                        self.client_message_handler(message, clients_message,
                                                    handler)
                    except Exception as e:
                        print(f'Error: {e}')
                        print(
                            f'Client {clients_message.getpeername()}'
                            f' was disconnected')
                        for user in self.users:
                            if self.users[user] == clients_message:
                                self.remove_user({'user': user},
                                                 clients_message)
                                break
                        if clients_message in self.clients:
                            self.clients.remove(clients_message)

    def add_users(self, user, message, sock):
        """Handler for adding authorized users to the shared dictionary"""

        client_ip, client_port = sock.getpeername()
        self.users[message['user']] = sock
        self.database.user_history(user, client_ip, client_port,
                                   status='login')
        self.database.add_active_users(user, client_ip, client_port)

    def remove_user(self, message, sock):
        """The handler of the client with which
        communication was interrupted.
        """

        client_ip, client_port = sock.getpeername()
        self.clients.remove(sock)
        self.users[message['user']].close()
        del self.users[message['user']]
        self.database.remove_active_users(message['user'])
        self.database.user_history(message['user'], client_ip, client_port,
                                   None)
        print(f'Client {client_ip}, {client_port} disconnected')

    @login_required
    def client_message_handler(self, message, client, handler):
        """Incoming message handler."""

        user = message['user']

        if message['action'] == 'presence':
            handler.message = {'OK': 'Server status OK: 200'}
            handler.send_message()

        elif message['action'] == 'registration':
            user = message['user']
            password = message['password']
            if self.database.create_user(user, password):
                handler.message = {'action': 'registered', 'user': user}
                handler.send_message()
                self.add_users(user, message, client)
            else:
                handler.message = {
                    'action': 'refuse',
                    'response': f'Sorry the name: {user}'
                                f' already exists.\n Please try again'
                }
                handler.send_message()

        elif message['action'] == 'authorisation':
            if self.database.check_username(user):
                handler.message = {
                    'action': None,
                    'response': f'Sorry the name: {user}'
                                f' is absent.\n Please try again'
                }
                handler.send_message()
            else:
                random_str = binascii.hexlify(os.urandom(64))
                server_hash = hmac.new(self.database.get_password(user),
                                       random_str, 'MD5')
                digest = server_hash.digest()
                handler.message = {
                    'user': user,
                    'data': random_str.decode('ascii'),
                    'response': 511
                }
                try:
                    handler.send_message()
                    handler.get_message()
                except OSError as e:
                    print(e)
                client_digest = binascii.a2b_base64(handler.message['data'])
                if handler.message['response'] == 511 and hmac.compare_digest(
                        digest, client_digest):
                    if user not in self.users.keys():
                        self.add_users(user, message, client)
                        handler.message = {
                            'action': 'success',
                            'user': user,
                            'response': 'User is authorized'
                        }
                        handler.send_message()
                else:
                    handler.message = {
                        'action': 'wrong_password',
                        'user': user,
                        'response': 'Wrong password. Please try again'
                    }
                    handler.send_message()

        elif message['action'] == 'update_key':
            self.database.update_key(user, message['key'])

        elif message['action'] == 'message':
            if message['addressee'] in self.users and self.users[
               message['addressee']] in self.messages:
                self.database.add_message_history(user, message['addressee'],
                                                  message['message'],
                                                  action='sent')
                response = self.database.message_count(user)
                handler.message = {
                    'action': 'sent_message',
                    'user': user,
                    'addressee': message['addressee'],
                    'count': response[0],
                    'message': message['message']
                }
                handler.send_message()

                self.database.add_message_history(user, message['addressee'],
                                                  message['message'],
                                                  action='')
                response = self.database.message_count(message['addressee'])
                handler.message = {
                    'action': 'message',
                    'user': user,
                    'addressee': message['addressee'],
                    'count': response[1],
                    'message': message['message']
                }
                handler.sock = self.users[message['addressee']]
                handler.send_message()
                handler.sock = client

            elif message['addressee'] in self.users and self.users[
                 message['addressee']] not in self.messages:
                print(
                    f'Connection with client {message["addressee"]} was lost')
            else:
                print(f'User {message["addressee"]} does not registered')

        elif message['action'] == 'get_contacts' and self.users[
             user] == client:
            response = self.database.get_contacts(user)
            if response:
                handler.message = {
                    'action': 'got_contacts',
                    'response': response
                }
                handler.send_message()
            else:
                handler.message = {
                    'action': 'no_contact',
                    'response': "You don't have any contact"
                }
                handler.send_message()

        elif message['action'] == 'add_contact' and self.users[user] == client:
            response = self.database.add_contact(user, message['contact'])
            if response:
                if response == 'no_contact':
                    handler.message = {
                        'action': 'no_contact',
                        'response': 'There is no such contact in the database'
                    }
                    handler.send_message()
                else:
                    handler.message = {
                        'action': 'added_contact',
                        'contact': message['contact'],
                        'response': f'The contact {message["contact"]}'
                                    f' was added'
                    }
                    handler.send_message()
            else:
                handler.message = {
                    'action': 'exist_contact',
                    'response': f'Sorry, the {message["contact"]}'
                                f' already exists'
                }
                handler.send_message()

        elif message['action'] == 'del_contact' and self.users[user] == client:
            response = self.database.del_contact(user, message['contact'])
            if response:
                if response == 'no_contact':
                    handler.message = {
                        'action': 'no_contact',
                        'response': 'There is no such contact'
                                    ' in your contact list'
                    }
                    handler.send_message()
                else:
                    handler.message = {
                        'action': 'deleted_contact',
                        'response': f'The contact {message["contact"]}'
                                    f' was deleted',
                        'contact': message['contact']
                    }
                    handler.send_message()
            else:
                handler.message = {
                    'action': 'no_contact',
                    'response': 'There is no such contact in the database'
                }
                handler.send_message()

        elif message['action'] == 'all_users':
            response = self.database.get_all_users()
            handler.message = {
                'action': 'got_all_users',
                'response': response
            }
            handler.send_message()

        elif message['action'] == 'get_contact_key':
            key = self.database.get_key(message['contact'])
            if key:
                handler.message = {
                    'action': 'got_contact_key',
                    'key': key,
                    'response': 511
                }
                handler.send_message()
            else:
                handler.message = {'response': 400}
                handler.send_message()

        elif message['action'] == 'exit':
            if user == 'Guest':
                self.clients.remove(client)
                print(f'Client {client.getpeername()} was disconnected')
            else:
                self.remove_user(message, client)
        else:
            print('Error: No action')



import binascii
import datetime
import hashlib

from sqlalchemy import create_engine, Table, Column, Integer, String,\
    MetaData, ForeignKey, DateTime, Text
from sqlalchemy.orm import mapper, Session
from sqlalchemy.sql import default_comparator


class Storage:
    """The class is a wrapper for working with the server's database.
     Uses SQLite database, implemented with
     SQLAlchemy ORM and uses the classic approach.
    """

    class Users:
        """Table class of all users."""

        def __init__(self, username, password):
            self.id = None
            self.username = username
            self.password = password
            self.key = None
            self.registration_date = datetime.datetime.now()

    class Contacts:
        """User's contact table class."""

        def __init__(self, user, contact):
            self.id = None
            self.user = user
            self.contact = contact
            self.add_contact_time = datetime.datetime.now()

    class UserHistory:
        """Login history table class."""

        def __init__(self, username, ip_address, port, login_time=None,
                     logout_time=None):
            self.id = None
            self.username = username
            self.login_time = login_time
            self.logout_time = logout_time
            self.ip_address = ip_address
            self.port = port

    class ActiveUsers:
        """Active user table class."""

        def __init__(self, user, ip_address, port, login_time):
            self.id = None
            self.user = user
            self.ip_address = ip_address
            self.port = port
            self.login_time = login_time

    class MessageHistory:
        """Message history table class."""

        def __init__(self, sender, recipient, sent_message, accepted_message):
            self.id = None
            self.sender = sender
            self.recipient = recipient
            self.sent_message = sent_message
            self.accepted_message = accepted_message
            self.data = datetime.datetime.now()

    class MessageCount:
        """Message counter table class."""

        def __init__(self, username):
            self.id = None
            self.username = username
            self.sent = 0
            self.accepted = 0

    def __init__(self):
        self.engine = create_engine('sqlite:///db/server_base.db3', echo=False,
                                    pool_recycle=7200,
                                    connect_args={'check_same_thread': False})
        self.metadata = MetaData()

        users = Table('Users', self.metadata,
                      Column('id', Integer, primary_key=True),
                      Column('username', String, unique=True),
                      Column('password', String),
                      Column('key', Text),
                      Column('registration_date', DateTime)
                      )

        contacts = Table('Contacts', self.metadata,
                         Column('id', Integer, primary_key=True),
                         Column('user', ForeignKey('Users.id')),
                         Column('contact', ForeignKey('Users.id')),
                         Column('add_contact_time', DateTime)
                         )

        user_history = Table('User_history', self.metadata,
                             Column('id', Integer, primary_key=True),
                             Column('username', ForeignKey('Users.id')),
                             Column('login_time', DateTime),
                             Column('logout_time', DateTime),
                             Column('ip_address', String),
                             Column('port', Integer)
                             )

        active_users = Table('Active_users', self.metadata,
                             Column('id', Integer, primary_key=True),
                             Column('user', ForeignKey('Users.id')),
                             Column('ip_address', String),
                             Column('port', Integer),
                             Column('login_time', DateTime)
                             )

        message_history = Table('Message_history', self.metadata,
                                Column('id', Integer, primary_key=True),
                                Column('sender', ForeignKey('Users.id')),
                                Column('recipient', ForeignKey('Users.id')),
                                Column('sent_message', String),
                                Column('accepted_message', String),
                                Column('data', DateTime)
                                )

        message_count = Table('Message_count', self.metadata,
                              Column('id', Integer, primary_key=True),
                              Column('username', String,
                                     ForeignKey('Users.id')),
                              Column('sent', Integer),
                              Column('accepted', Integer)
                              )

        self.metadata.create_all(self.engine)

        mapper(self.Users, users)
        mapper(self.Contacts, contacts)
        mapper(self.UserHistory, user_history)
        mapper(self.ActiveUsers, active_users)
        mapper(self.MessageHistory, message_history)
        mapper(self.MessageCount, message_count)

        self.session = Session(bind=self.engine)
        self.session.query(self.ActiveUsers).delete()
        self.session.commit()

    def check_username(self, username):
        """Database user check."""

        if self.session.query(self.Users).filter_by(username=username).count():
            return False
        else:
            return True

    def get_password(self, username):
        """Get user password."""

        password = self.session.query(self.Users.password).filter_by(
            username=username)
        return password.one()[0]

    def create_user(self, username, password):
        """Creating a new user and entering into the database"""

        if self.check_username(username):
            password_bytes = password.encode('utf-8')
            solt_bytes = username.encode('utf-8')
            password_hash = hashlib.pbkdf2_hmac('sha256', password_bytes,
                                                solt_bytes, 10000)
            new_user = self.Users(username, binascii.hexlify(password_hash))
            self.session.add(new_user)
            self.session.commit()
            new_user_in_message_count = self.MessageCount(username)
            self.session.add(new_user_in_message_count)
            self.session.commit()
            return True
        else:
            return False

    def add_active_users(self, username, ipaddress, port):
        """Adding an active user to the database table"""

        user = self.ActiveUsers(username, ipaddress, port,
                                login_time=datetime.datetime.now())
        self.session.add(user)
        self.session.commit()

    def update_key(self, username, key):
        """Update user's key"""

        user_key = self.session.query(self.Users).filter_by(
            username=username).first()
        if user_key.key != key:
            user_key.key = key

    def remove_active_users(self, username):
        """Removing an outgoing user"""

        self.session.query(self.ActiveUsers).filter_by(user=username).delete()
        self.session.commit()

    def user_history(self, username, ip_address, port, status):
        """Adding user inputs and outputs to the database table"""

        if status == 'login':
            history = self.UserHistory(username, ip_address, port,
                                       login_time=datetime.datetime.now())
            self.session.add(history)
            self.session.commit()
        else:
            user = self.session.query(self.UserHistory).filter_by(
                username=username, logout_time=None) \
                .order_by(self.UserHistory.id.desc()).first()
            setattr(user, 'logout_time', datetime.datetime.now())
            self.session.commit()

    def get_key(self, username):
        """Getting a user key"""

        key = self.session.query(self.Users).filter_by(
            username=username).first()
        return key.key

    def get_contacts(self, username):
        """Getting a user's contact list"""

        return [user[0] for user in
                self.session.query(self.Contacts.contact).filter_by(
                    user=username).all()]

    def add_contact(self, user, contact):
        """Adding a user to the selected user's contact list"""

        if self.check_username(contact) is False:
            if self.session.query(self.Contacts).filter_by(user=user,
                                                           contact=contact).\
                                                                    count():
                return False
            else:
                add_contact = self.Contacts(user, contact)
                self.session.add(add_contact)
                self.session.commit()
                return True
        else:
            return 'no_contact'

    def del_contact(self, user, contact):
        """Removing a user from the selected user's contact list"""

        if self.check_username(contact) is False:
            if self.session.query(self.Contacts).filter_by(user=user,
                                                           contact=contact).\
                                                                    count():
                self.session.query(self.Contacts).filter_by(user=user,
                                                            contact=contact).\
                                                                    delete()
                self.session.commit()
                return True
            else:
                return 'no_contact'
        else:
            return False

    def active_users_list(self):
        """Getting a list of active users"""

        query = self.session.query(
            self.ActiveUsers.user,
            self.ActiveUsers.ip_address,
            self.ActiveUsers.port,
            self.ActiveUsers.login_time
        )
        return query.all()

    def get_all_users(self):
        """Getting all known users"""

        return [user[0] for user in
                self.session.query(self.Users.username).all()]

    def message_count(self, username):
        """Get the number of messages of the selected user"""

        query = self.session.query(
            self.MessageCount.sent,
            self.MessageCount.accepted
        ).filter_by(username=username).all()
        return query[0]

    def add_message_history(self, sender, recipient, message, action):
        """Adding message history to the database table of the selected user"""

        if action == 'sent':
            row = self.MessageHistory(sender, recipient, message,
                                      accepted_message=None)
            self.session.add(row)
            sender_row = self.session.query(self.MessageCount).filter_by(
                username=sender).first()
            sender_row.sent += 1
            self.session.commit()
        else:
            row = self.MessageHistory(sender, recipient, sent_message=None,
                                      accepted_message=message)
            self.session.add(row)
            recipient_row = self.session.query(self.MessageCount).filter_by(
                username=recipient).first()
            recipient_row.accepted += 1
            self.session.commit()


if __name__ == '__main__':
    test = Storage()
    test.message_count('qqq')



from socket import socket


def login_required(func):
    """A decorator that checks that the client is authorized on the server.
    Verifies that the socket object being passed is in
    list of authorized clients.
    Except for passing a query dictionary
    for authorization. If the client is not authorized,
    throws a TypeError exception
    """

    def checker(*args, **kwargs):
        from core import Server

        if isinstance(args[0], Server):
            found = False
            for arg in args:
                if isinstance(arg, socket):
                    for client in args[0].users:
                        if args[0].users[client] == arg:
                            found = True
            for arg in args:
                if isinstance(arg, dict):
                    if 'action' in arg and arg['action'] == 'presence'\
                            or arg['action'] == 'authorisation'\
                            or arg['action'] == 'registration':
                        found = True
            if not found:
                raise TypeError
        return func(*args, **kwargs)
    return checker



from PyQt5.QtWidgets import QMainWindow, QAction, qApp, QLabel, QTableView
from PyQt5.QtGui import QStandardItemModel, QStandardItem
# from PyQt5.QtCore import QTimer
# from server.statistic import StatWindow
from config_window import ConfigWindow


class MainWindow(QMainWindow):
    """Main window of server"""

    def __init__(self, database, config):
        super().__init__()

        self.database = database

        self.config = config

        self.exitAction = QAction('Exit', self)
        self.exitAction.setShortcut('Ctrl+Q')
        self.exitAction.triggered.connect(qApp.quit)

        self.refresh_button = QAction('Update list', self)

        self.config_btn = QAction('Server setting', self)

        # self.show_history_button = QAction('User history', self)

        self.statusBar()
        self.statusBar().showMessage('Server Working')

        self.toolbar = self.addToolBar('MainBar')
        self.toolbar.addAction(self.exitAction)
        self.toolbar.addAction(self.refresh_button)
        # self.toolbar.addAction(self.show_history_button)
        self.toolbar.addAction(self.config_btn)

        self.setFixedSize(800, 600)
        self.setWindowTitle('Messaging Server alpha release')

        self.label = QLabel('User connection list:', self)
        self.label.setFixedSize(240, 15)
        self.label.move(10, 25)

        self.active_clients_table = QTableView(self)
        self.active_clients_table.move(10, 45)
        self.active_clients_table.setFixedSize(780, 400)

        # self.timer = QTimer()
        # self.timer.timeout.connect(self.create_users_model)
        # self.timer.start(1000)

        self.refresh_button.triggered.connect(self.create_users_model)
        # self.show_history_button.triggered.connect(self.show_statistics)
        self.config_btn.triggered.connect(self.server_config)

        self.show()

    def create_users_model(self):
        """The method that fills is a table of active users."""

        list_users = self.database.active_users_list()
        user_list = QStandardItemModel()
        user_list.setHorizontalHeaderLabels(
            ['username', 'IP address', 'Port', 'login_time'])
        for row in list_users:
            user, ip, port, login_time = row
            user = QStandardItem(user)
            user.setEditable(False)
            ip = QStandardItem(ip)
            ip.setEditable(False)
            port = QStandardItem(str(port))
            port.setEditable(False)
            login_time = QStandardItem(str(login_time.replace(microsecond=0)))
            login_time.setEditable(False)
            user_list.appendRow([user, ip, port, login_time])
        self.active_clients_table.setModel(user_list)
        self.active_clients_table.resizeColumnsToContents()
        self.active_clients_table.resizeRowsToContents()

    # def show_statistics(self):
    #     global stat_window
    #     stat_window = StatWindow(self.database)
    #     stat_window.show()

    def server_config(self):
        """Method that creates a window with server settings."""

        global config_window
        config_window = ConfigWindow(self.config)



import json
import sys

from descripts import SetName


class Handler:
    """Main class responsible for sending and receiving messages"""

    username = SetName()

    def __init__(self):
        self.message = {}
        self.username = 'Guest'
        self.sock = None

    def send_message(self):
        json_message = json.dumps(self.message)
        self.sock.send(json_message.encode('utf-8'))

    def get_message(self):
        encoded_response = self.sock.recv(1024)
        if isinstance(encoded_response, bytes):
            json_response = encoded_response.decode('utf-8')
            if not json_response:
                sys.exit()
            response = json.loads(json_response)
            if isinstance(response, dict):
                self.message = response
            else:
                raise Exception
        else:
            raise Exception



from setuptools import setup, find_packages

setup(name="fine_messanger_server",
      version="0.0.1",
      description="server",
      author="Vyacheslav",
      author_email="slavkin.131@gmail.com",
      packages=find_packages(),
      install_requires=['PyQt5', 'sqlalchemy', 'pycryptodome', 'pycryptodomex']
      )



import ipaddress


class CheckPort:
    """Handle responsible for checking the correct port"""

    def __set_name__(self, owner, name):
        self.name = name

    def __set__(self, instance, value):

        if not 1023 < value < 65536:
            raise ConnectionError(f'Incorrect port "{value}".'
                                  f' The port must be between 1023 and 65536.')
        instance.__dict__[self.name] = value


class CheckIp:
    """Handle responsible for checking the correct ip address"""

    def __set_name__(self, owner, name):
        self.name = name

    def __set__(self, instance, value):
        if value == '':
            value = '127.0.0.1'
        try:
            ipaddress.ip_address(value)
            instance.__dict__[self.name] = value
        except ValueError as e:
            print(f'Error: {e}')
            exit(1)


class SetName:
    """Handle responsible for checking the correct username"""

    def __set_name__(self, owner, name):
        self.name = name

    def __set__(self, instance, value):
        if 1 < len(value) < 20:
            instance.__dict__[self.name] = value
        else:
            instance.__dict__[self.name] = None



import configparser
import os
import sys

from PyQt5.QtWidgets import QApplication

from common.parser import Parser
from server.core import Server
from server.database import Storage
from server.server_gui import MainWindow


def main():
    """Server start function"""

    config = configparser.ConfigParser()
    dir_path = os.getcwd()
    config.read(f"{dir_path}/{'server.ini'}")

    connection = Parser()
    connection.arg_parser()
    ip_address = connection.ip_address
    port = connection.port
    database = Storage()
    server = Server(ip_address, port, database)
    server.daemon = True
    server.start()

    server_app = QApplication(sys.argv)
    main_window = MainWindow(database, config)
    server_app.exec_()


if __name__ == '__main__':
    main()



import argparse
import sys


class Parser:
    """Command line argument parser"""

    ip_address = ''
    port = 7777
    client_name = 'Guest'

    def arg_parser(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('-p', default=self.port, type=int, nargs='?')
        parser.add_argument('-a', default=self.ip_address, nargs='?')
        parser.add_argument('-n', '--name', default=self.client_name,
                            nargs='?')
        namespace = parser.parse_args(sys.argv[1:])
        self.ip_address = namespace.a
        self.port = namespace.p
        self.client_name = namespace.name



import dis


class CheckStatus(type):
    def __init__(cls, name, base, attrs):
        super().__init__(name, base, attrs)
        print(f'{name} started...')
        load_global = []
        for i in attrs:
            try:
                ret = dis.get_instructions(attrs[i])
            except TypeError:
                pass
            else:
                for j in ret:
                    if j.opname == 'LOAD_GLOBAL':
                        if j.argval not in load_global:
                            load_global.append(j.argval)
        # print(load_global)

        if name == 'Server':
            if 'connect' in load_global:
                raise TypeError('Using connect method is not allowed'
                                ' in server class')
            if not ('SOCK_STREAM' in load_global and 'AF_INET' in load_global):
                raise TypeError('Incorrect socket initialization')

        elif name == 'Client':
            if 'accept' in load_global or 'listen' in load_global:
                raise TypeError('The use of a forbidden method was detected'
                                ' in the class')
            if not ('SOCK_STREAM' in load_global and 'AF_INET' in load_global):
                raise TypeError('Incorrect socket initialization')
